from .robot import Robot
from .direction import Direction


import logging
logging.basicConfig(level=logging.DEBUG, format='(%(threadName)-9s) %(message)s',)


''' control robot positioning and drawing '''
class RobotPosition( Robot ):

    def __init__( self, level, **kwargs ):
      self.grid  = level.draw_grid
      super().__init__( level, **kwargs )

      # the position in pixels
      self.x = 0
      self.y = 0

      self.step = 4
      self.robot_size = 64

      if self.maze is None:
          self.x_size = self.grid.width_pixels
          self.y_size = self.grid.height_pixels
      else:
          x,y = self.maze.dimensions()
          self.x_size = x * self.robot_size
          self.y_size = y * self.robot_size


    def get_cell_position(self):
      ''' get the current position in grid coords '''
      return self.x//self.robot_size, self.y//self.robot_size


    def set_cell_position(self, *args):
      ''' set the robot position in grid coords '''
      # convert from grid position to pixels
      if len(args) == 1:
        self.x,self.y = self.grid.grid_to_pixels(args[0])
        self.x_cell = args[0][0]
        self.y_cell = args[0][1]
      elif len(args) == 2:
        self.x,self.y = self.grid.grid_to_pixels([args[0],args[1]])
        self.x_cell = args[0]
        self.y_cell = args[1]


    def move(self,new_x,new_y):
        ' move from the current position to the specified position '

        if (self.x_cell == new_x) and (self.y_cell == new_y):
           self.move_direction(Direction.Stay)

        while (self.x_cell < (self.level.width-1)) and (self.x_cell < new_x):
          self.move_direction(Direction.East)
          self.x_cell += 1

        while (self.x_cell > 0) and (self.x_cell > new_x):
          self.move_direction(Direction.West)
          self.x_cell -= 1

        while (self.y_cell < (self.level.height-1)) and (self.y_cell < new_y):
          self.move_direction(Direction.South)
          self.y_cell += 1

        while (self.y_cell > 0) and (self.y_cell > new_y):
          self.move_direction(Direction.North)
          self.y_cell -= 1


    def move_East(self):
        if self.x < (self.x_size - self.robot_size):
            self.x += self.step

    def move_West(self):
        if self.x > 0:
            self.x -= self.step

    def move_North(self):
        if self.y > 0:
            self.y -= self.step

    def move_South(self):
        if self.y < (self.y_size - self.robot_size):
            self.y += self.step


import numpy as np
import os
from typing import Union

from .direction import Direction
from .grid_base import GridBase
from .grid_info import GridInfo
from .draw_grid import DrawGrid
from .draw_info import DrawInfo



class GridLevel():
  '''
    Class to manage a grid level.
    This performs all drawing and querying of the level
    (to see which moves are possible within a given grid cell)
  '''

  def __init__( self, **kwargs: dict ):

    # get the directory where this file is running
    dir_path = os.path.dirname(os.path.realpath(__file__))

    self.grid_base = GridBase( dir_path, **kwargs )
    self.grid_info = GridInfo( self.grid_base, **kwargs )
    self.draw_grid = DrawGrid( self.grid_base, **kwargs )
    self.draw_info = DrawInfo( self.draw_grid, self.grid_info, **kwargs )


  '''
      Query Functions
  '''

  def get_directions( self, x: int = None, y: int = None ) -> Union[Direction,np.ndarray]:
    ''' return the possible directions for the specified grid cell
        - if a grid cell is not specified then return an array of possible
        directions for all grid cells
    '''
    return self.grid_info.get_directions( x, y )


  def get_rewards( self, x: int = None, y: int = None ) -> Union[Direction,np.ndarray]:
    ''' return the reward for the specified grid cell
        - if a grid cell is not specified then return an array of rewards for all grid cells
    '''
    return self.grid_base.get_reward( x, y )


  def get_next_state( self, x, y, direction ):
    ''' return the next state and reward for moving
        - (x,y) = current state
        - direction = direction moved from current state
    '''

    assert direction >= Direction.Stay and direction <= Direction.West

    # check that some actions are possible in this state
    possible_actions = self.grid_info.get_cell_directions(x,y,direction)
    if not possible_actions:
      # stay in same position, reward = -1 for trying to move
      # - target only reached if choosing to stay in same state
      return [x,y],-1,(direction==Direction.Stay)

    # a deterministic policy should only have one possible action
    chosen_action = [key for (key, value) in possible_actions.items() if value]
    if len(chosen_action) != 1:
      # stay in same position, reward = -1 for trying to move
      return [x,y],-1,False

    # get the list of all other possible states
    all_actions = self.grid_info.get_cell_directions(x,y)
    all_actions.pop(chosen_action[0], None)
    other_states = [key for (key, value) in all_actions.items() if value]

    # get the probability of moving to the intended target
    transition_probability = self.grid_base.get_transition_probability( x, y )

    # if the probability is less than the transition probability then move to the target
    # of if the target state is the only allowed state
    target_state_reached = True
    if (np.random.random() < transition_probability) or (len(other_states) == 0):
      direction = chosen_action[0]
    else:
      # choose one of the other possible states
      direction = np.random.choice(other_states)

      # set the flag to indicate the target state wasn't reached
      target_state_reached = False

    # calculate the postion of the next state
    next_pos = self.get_next_state_position( x, y, direction )

    # get the reward for taking this action
    reward = self.grid_base.get_reward( next_pos[0], next_pos[1] )

    # for equal probability of taking an action its just the mean of all actions
    return next_pos, reward, target_state_reached


  def get_next_state_position( self, x, y, direction ):
    ''' given the current state position and direction calculate the postion of the next state '''
    next_pos = []
    if direction == 'N': next_pos = [x,y-1]
    if direction == 'S': next_pos = [x,y+1]
    if direction == 'E': next_pos = [x+1,y]
    if direction == 'W': next_pos = [x-1,y]
    return next_pos


  def get_reward( self, x, y, direction = None ):
    ''' get the reward for moving to the cell at (x,y) or, if a direction is specified,
        the reward for moving from the cell at (x,y) to the cell in the specified direction
    '''
    if direction:

      # if the direction is given as an enum convert to char
      if type(direction) != str:
        direction = Direction.get_direction_char(direction)

      # calculate the postion of the next state
      nx,ny = self.get_next_state_position( x, y, direction )
      return self.grid_base.get_reward(nx,ny),[nx,ny]

    # get the reward for taking this action
    return self.grid_base.get_reward(x,y)


  '''
      Graphical Functions
  '''

  def draw( self ):
    ''' render the grid canvases '''
    return self.draw_grid.canvases

  def clear( self, all_info=False ):
    ''' clear anything currently in the info panels '''
    self.draw_grid.clear(all_info)

  def show_info( self, info: dict ):
    ''' add the supplied information to the grid '''
    self.draw_info.draw( info )

  def save( self, filename ):
    ''' render the grid canvases '''
    canvases = self.get_canvases()
    # save and restore to complete any canvas drawing
    canvases[3].save()
    canvases[3].restore()
    return canvases.to_file(filename)

  def get_canvases(self):
    ''' get the grid levels multi-canvas '''
    return self.draw_grid.canvases

  def get_canvas_dimensions( self ):
    ''' get the total size of the grid canvas in pixels '''
    return [self.draw_grid.total_width, self.draw_grid.total_height]



from typing import List
from ..envs import BabyRobotInterface
from ..envs.lib.actions import Actions
from .policy import Policy


class DeterministicPolicy(Policy):
  ''' a really simple deterministic policy class
      - given a list of actions it will simply return the next action when the state changes
  '''
        
  def __init__(self, level: BabyRobotInterface, actions: List[Actions]):    
    super().__init__( level )    
    self.actions = actions
    self.action_index = -1
    self.last_x = -1
    self.last_y = -1
        
  def get_action(self,x:int,y:int) -> Actions:  
    if x != self.last_x or y != self.last_y:
        self.action_index += 1
        if self.action_index > len(self.actions): self.action_index = 0
        self.last_x = x
        self.last_y = y
    return self.actions[self.action_index]


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.


import numpy as np
from numpy import inf

from babyrobot.envs import BabyRobotInterface
from babyrobot.envs.lib.direction import Direction


class ValueIteration():

  def __init__(self, env: BabyRobotInterface, discount_factor=0.9):
    self.level = env
    self.values = np.zeros((env.height,env.width))
    self.discount_factor = discount_factor


  def get_state_value(self,pos):
    ''' get the currently calculated value of the specified position in the grid '''
    x = pos[0]
    y = pos[1]
    if (x < 0 or x >= self.level.width) or (y < 0 or y >= self.level.height): 
      return 0
    return self.values[y,x]


  def calculate_max_action_value(self,x,y):
    ''' 
        calculate the values of all actions in the specified cell and return the largest of these
        - the next state value is given by:  v(s) = max[r + γv(s')]       
    '''    
    # get the probability of moving to the intended target from this state
    transition_probability = self.level.get_transition_probability( x, y )

    # get the list of all possible actions in this state
    all_actions = self.level.get_available_actions(x,y)

    # when an action is taken will either end in target state or one of the other possible states
    # - count the number of states other than the target state
    num_alternative_states = len(all_actions) - 1 

    # calculate the value of each action in the state and save the largest
    max_value = float('-inf')
    for chosen_action in all_actions:

      # sum the values of the possible next states
      value = 0  
      for action in all_actions:

        if action == chosen_action:
          # the chosen action is taken with the cell's transition probability
          probability = transition_probability
        else:
          # the probability of ending up in another state is divided by the total number of other possible states
          probability = (1-transition_probability)/num_alternative_states

        # get the reward for moving from the current cell in this direction
        reward, next_pos = self.level.get_reward( x, y, Direction.from_action(action) )   

        # combine the reward with discounted value of the next state and 
        # sum over each of the transition probabilities p(s'|s,a)
        value += probability * (reward + (self.discount_factor * self.get_state_value( next_pos ))) 

      # save the largest value
      if value > max_value:
        max_value = value              
    
    # return the value of the largest action-value in this state
    return max_value


  def state_sweep(self):
    ''' calculate the value of all states except the exit '''
    
    new_values = np.zeros((self.level.height,self.level.width))
    end = self.level.end
    for y in range(self.level.height):
      for x in range(self.level.width):
        if (x != end[0]) or (y != end[1]):          
          new_values[y,x] = self.calculate_max_action_value(x,y)    

    # calculate the largest difference in the state values between the start and end of the sweep    
    delta = np.max(np.abs(new_values - self.values))           
    
    # update the state values with the newly calculated values 
    # (set any -inf to NaN for states where no value calculated)
    new_values[new_values == -inf] = np.nan
    self.values = new_values

    # return the largest state value difference 
    return delta


  def run_to_convergence(self, max_iterations = 100, threshold = 1e-3):
    ''' run multiple state sweeps until the maximum change in the state value falls
        below the supplied threshold or the maximum number of iterations is reached
    '''    
    for n in range(max_iterations):
      
      # calculate the maximum action value in each state and get the largest state value difference
      delta = self.state_sweep()        
      
      # test if the difference is less than the defined convergence threshold
      if delta < threshold:
        break
    
    # return the number of iterations taken to converge
    return n


from argparse import Action
import gymnasium
import numpy as np
from .actions import Actions

class Dynamic(gymnasium.Space):

  def __init__(self, action_list = []):
      ' set the list of initially available actions '
      self.set_actions(action_list)

  def sample(self):
      ' select a random action from the set of available actions '
      if len(self.available_actions) > 0:
        return np.random.choice(self.available_actions)
      return Actions.Stay

  def set_actions(self,actions):
      self.available_actions = actions
      self.n = len(actions)

  def get_available_actions(self):
      return [str(action) for action in self.available_actions]


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

import math
import numpy as np
from ..envs import BabyRobotInterface
from ..envs.lib.direction import Direction
from ..envs.lib.actions import Actions

class Policy():

  def __init__(self, level: BabyRobotInterface, directions: np.array = None, seed = None):
    self.level = level
    if directions is None:
      # if no directions are supplied create a policy where all actions are equally likely
      self.directions = np.full((level.height,level.width), Direction.All)
    else:
      self.directions = directions

    # no directions are possible in the terminal state
    self.directions[level.end[1],level.end[0]] = 0

    # set the seed used to choose random actions for a stochastic policy
    np.random.seed(seed=seed)

  def set_policy(self,directions):
    ''' set the policy (i.e. the action to take in each state) '''
    self.directions = directions

  def get_policy(self):
    return self.directions

  def get_directions(self,values = None):
    if values is not None:
      self.directions = self.calculate_greedy_directions(values)
    return self.directions

  def update_policy(self,values):
    ''' act greedily wrt to the supplied state values to get a new set of actions
      - if the new greedy policy specifies more than one action for a state select
        the action that was in previous policy
    '''
    greedy_directions = self.calculate_greedy_directions(values)

    for row in range(self.directions.shape[0]):
      for col in range(self.directions.shape[1]):
        # if a single direction is specified the value will be a power of 2
        n = greedy_directions[row,col]
        power_of_two = (n & (n-1) == 0) and n != 0
        if not power_of_two:
          # more than one direction so use direction from last policy
          greedy_directions[row,col] = self.directions[row,col]

    # update the policy
    self.directions = greedy_directions
    return self.directions

  def calculate_greedy_directions(self,values):
    ''' given a set of state values calculate the directions by acting greedily
        i.e. move in the direction of greatest state value
    '''
    # calculate the directions of all states except the exit
    directions = np.zeros((self.level.height,self.level.width),dtype=int)
    end = self.level.end
    for y in range(self.level.height):
      for x in range(self.level.width):
        if (x != end[0]) or (y != end[1]):
          if len(values.shape) == 2:
            directions[y,x] = self.calculate_cell_directions(x,y,values)
          else:
            directions[y,x] = self.get_greedy_direction(values[y][x])
    return directions


  def get_greedy_direction( self, arr ):
    ''' return the direction(s) with the maximum action value '''
    directions = 0
    best_value = -np.inf
    for action in range(len(Actions)):
      dir_value = Direction.from_action(action) 
      action_value = arr[action]
      if directions > 0 and math.isclose( action_value, best_value, rel_tol=1e-6):
        directions += dir_value
      elif action_value != 0 and action_value > best_value:
        directions = dir_value
        best_value = action_value 
    return directions 


  def calculate_cell_directions(self,x,y,values):
    ''' select the action with the highest value
        = argmax[ sum(p(s',r|s,a)[r + γV(s')]) ]
    '''
    actions = self.level.get_available_actions(x,y)

    directions = 0
    dir_value = 0
    best_value = -np.inf
    for action in actions:
      # get the possible probabilities, next_states and rewards for this action
      action_probabilities = self.level.get_action_probabilities( x, y, action )

      action_value = 0
      for probability,next_state,reward in action_probabilities:
        # convert the x,y position into row,col
        action_value += probability * (reward + values[next_state[1],next_state[0]])

      dir_value = Direction.from_action(action)

      # if a best action has already been selected and the new action has a value
      # very close to this, then add this action to the set of greedy actions
      if directions > 0 and math.isclose( action_value, best_value, rel_tol=1e-6):
        directions += dir_value
      elif action_value > best_value:
        directions = dir_value
        best_value = action_value

    return int(directions)


  def get_state_directions(self,x,y):
    ''' return the direction bitfield for the specified state
      - this combines the directions allowed by the grid with those specified by the policy
    '''
    # get the list of all possible actions in this state
    cell_actions = self.level.get_available_actions(x,y)

    # convert the actions to a direction bitfield
    cell_directions = Direction.from_actions( cell_actions )

    # do a bitfield combination of the cell's allowed directions with the policy
    #  to get the available directions
    return (cell_directions & self.directions[y,x])


  def get_direction_list(self,x,y):
    ''' return a list of the allowed directions for the specified state
      - this combines the directions allowed by the grid with those specified by the policy
    '''
    return Direction.get_list( self.get_state_directions(x,y) )


  def get_actions(self,x,y):
    ''' return a list of the actions specified by the policy for the specified state
      - this takes the actions specified by the policy and removes any that are invalid
      (i.e. any that aren't possible in the environment)
    '''
    direction_value = self.get_state_directions(x,y)
    action_list = []
    if direction_value & Direction.North: action_list.append( Actions.North )
    if direction_value & Direction.South: action_list.append( Actions.South )
    if direction_value & Direction.East:  action_list.append( Actions.East )
    if direction_value & Direction.West:  action_list.append( Actions.West )
    return action_list


  def get_action(self,x,y):
    ''' return a single action for the specified state
        - if no action exists for this state then one will be chosen at random from the
        available directions
        - if more than one action exists then one will be chosen from these at random
    '''
    actions = self.get_actions(x,y)
    if len(actions) == 0:
      # choose a random action (stochastic policy with all actions possible)
      return self.level.action_space.sample()

    # choose one of the policies possible actions
    return np.random.choice(actions)


  def get_action_probabilities(self,x,y):
    ''' return a dictionary with the allowed actions for a state along with the
        probability of taking each of these actions
    '''
    # get the allowed actions in the state
    actions = self.get_actions(x,y)

    if len(actions) == 0:
      return {}
    else:
      equal_probability = 1 / len(actions)
      return { action:equal_probability for action in actions }


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

import os
import numpy as np
import imageio.v2 as imageio

from ..envs import BabyRobotInterface
from ..envs.lib import Actions
from ..envs.lib import Direction
from . import Utils
from . import PolicyEvaluation
from . import Policy

from ipywidgets import HBox, VBox

from time import sleep
import os
import time

import logging
logging.getLogger().setLevel(logging.ERROR)


class Animate():


  def __init__(self, env: BabyRobotInterface):

    self.env = env
    self.canvases = env.level.get_canvases()

    self.step = 0
    self.partial_step = 0
    self.max_partial_step = 0

    # the default time between each step when just displaying the episode
    self.kPlayInterval = 80

    # if set true an image will be written for each step of the animation
    self.create_images: False

    # the default time between each step when creating images from the episode
    self.kImageInterval = 300


  def set_parameters(self, **kwargs):
    '''
        set all the parameters using any supplied values,
        otherwise use the defaults
    '''
    # the maximum number of steps for which the episode should run
    self.max_steps = kwargs.get('max_steps',10)

    # the number of iterations between each displayed update or image write
    self.save_interval = kwargs.get('save_interval',1)

    # the precision to display the state values
    self.precision = kwargs.get('precision',1)

    # test if directions should be displayed
    self.show_directions = kwargs.get('show_directions',True)

    # test if an image should be written at each step
    self.create_images = kwargs.get('create_images',False)

    # the folder to use when generating images during the run
    self.image_folder = kwargs.get('image_folder',"animate_images")

    self.play_interval = kwargs.get('interval',self.kPlayInterval)
    if self.create_images and self.play_interval < self.kImageInterval:
      # when images are being written need to go more slowly
      self.play_interval = self.kImageInterval


  def get_info_string( self, details ):
    '''
        basic function to display text information
    '''
    info = {}
    info['side_info'] = \
    [
      ((10,10),f"step: {details['step']}"),
      ((10,30),f"state: {details['state']}"),
      ((10,50),f"action: {details['action']}"),
      ((10,70),f"new state: {details['new_state']}"),
      ((10,90),f"direction: {details['direction']}"),
      ((10,110),f"reward: {details['reward']}"),
      ((10,130),f"done: {details['done']}"),
      ((10,170),f"total reward: {details['total_reward']}"),
    ]
    return info


  def write_file(self):
    '''
        write the current environment to a file and wait for it to be created
    '''
    filename = f'{self.image_folder}/step_{self.partial_step}.png'

    # save the environment to a file
    self.canvases[3].save()
    self.canvases[3].restore()
    self.canvases.to_file(filename)

    # wait for the file to be created
    time_to_wait = 10
    time_counter = 0
    while not os.path.exists(filename):
        time.sleep(1)
        time_counter += 1
        if time_counter > time_to_wait:break
    time.sleep(5)


  def clear_image_folder(self):
    '''
        Helper method to delete any images in the image folder
    '''
    Utils.clear_directory(self.image_folder)


  def create_movie( self, **kwargs ):
    '''
        create a gif movie from the images in the specified directory
        where:
          duration = time between each frame
    '''
    movie_name = kwargs.get('movie_name',"")
    movie_frames = kwargs.get('movie_frames',self.max_partial_step)

    # use the supplied duration parameter if present
    if 'duration' in kwargs:
      self.duration = kwargs.get('duration')

    print(f"Creating the movie: {movie_name} (duration = {self.duration}, steps = {movie_frames})",end="")
    with imageio.get_writer(movie_name, mode='I', duration=self.duration) as writer:
      for index in range(0,movie_frames+1):
        file = f"{self.image_folder}/step_{index}.png"
        if os.path.exists(file):
          image = imageio.imread(file)
          writer.append_data(image)
    print(f" - Complete")

    # remove the images when the movie has been made
    if kwargs.get('clear_images',True):
      self.clear_image_folder()


  def show(self, on_update, **kwargs):
    '''
        the base show function that performs the common operations of all show functions
    '''
    self.env.reset()
    self.total_reward = 0
    self.step = 0
    self.partial_step = 0
    self.done = False

    # define a callback function to call when the canvases are changed
    # - this then saves the current view
    def save_to_file(*args, **kwargs):
        # do a save and restore to force canvas update before writing
        self.canvases[3].save()
        self.canvases[3].restore()
        self.canvases.to_file(f'{self.image_folder}/step_{self.partial_step}.png')

    if self.create_images:
      Utils.create_image_directory(self.image_folder)
      self.canvases.observe(save_to_file,'image_data')

    # run multiple policy iterations
    play, progress, layout = Utils.setup_play_level( self.env.level, on_update, max=(self.max_partial_step/self.save_interval), interval=self.play_interval )
    return VBox((self.canvases, HBox((play, progress))),layout=layout)


  def show_policy(self, policy: Policy, **kwargs):
    '''
        animate the steps of the supplied policy
    '''
    # store the supplied policy
    self.policy = policy

    # the default time between each step when just displaying the episode
    self.kPlayInterval = 80

    # the default time between each step when creating images from the episode
    self.kImageInterval = 300

    # setup any supplied parameter values
    self.set_parameters(**kwargs)

    # the time between each frame if a movie is created
    self.duration = 0.08

    # test if information should be shown
    # - the default function relies on the side panel being created for this
    info_function = kwargs.get('info_function',None)
    if kwargs.get('show_info') and info_function is None:
      # if an info function has not been supplied use the basic class member function
      info_function = self.get_info_string

    # each step is broken into 16 step_interval
    self.max_steps = kwargs.get('max_steps',10)
    step_interval = 16
    self.max_partial_step = (self.max_steps * step_interval)

    self.last_action = self.policy.get_action(self.env.x,self.env.y)
    self.last_state = np.array([0,0])

    self.current_state = [self.env.x,self.env.y]
    self.new_state = [self.env.x,self.env.y]

    def on_update(*args):

      if self.partial_step < self.max_partial_step:

        if (self.partial_step % step_interval) == 0:

          self.step += 1

          # keep going until the exit is reached
          if not self.done:

            self.current_state = [self.env.x,self.env.y]

            # get an action from the environment
            action = self.policy.get_action(self.env.x,self.env.y)


            # take the action and get the properties of the next state
            self.new_state, reward, self.done, truncated, info = self.env.step(action)
            self.total_reward += reward

            if info_function is not None:
              details = \
              {
                'state': self.current_state,
                'action': Direction.from_action(action),
                'reward': reward,
                'new_state': self.new_state,
                'total_reward': self.total_reward,
                'done': self.done,
                'truncated': truncated,
                'info': info,
                'step': self.step,
                'direction': Direction.get_string(Direction.get_direction(self.current_state,self.new_state))
              }
              self.env.show_info(info_function(details))

            # if the state hasn't changed set the last action to be 'stay'
            self.last_action = action
            if np.array_equal(self.new_state,self.last_state):
              self.last_action = Actions.Stay
            self.last_state = self.new_state

          else:
            # set the last action to 'stay' to signify the exit being reached
            self.last_action = Actions.Stay

        # test if an action that caused a move occurred
        if Actions(self.last_action) is not Actions.Stay:

          direction = Direction.get_direction(self.current_state,self.new_state)

          # 'partial_move' moves part of the way to the next state
          self.env.robot.partial_move(direction)

        self.partial_step += 1

      else:
        # reset at the end of the run
        self.last_state = [self.env.x,self.env.y]
        self.partial_step = 0
        self.done = False

    return self.show(on_update,**kwargs)


  def show_policy_evaluation(self, policy_evaluation: PolicyEvaluation, **kwargs):
    '''
        animate the iterations of policy evaluation
    '''
    # store the supplied policy evaluation object
    self.policy_evaluation = policy_evaluation

    # the default time between each step when just displaying the episode
    self.kPlayInterval = 500

    # the default time between each step when creating images from the episode
    self.kImageInterval = 3000

    # setup any supplied parameter values
    self.set_parameters(**kwargs)

    # the time between each frame if a movie is created
    self.duration = 1.0

    # each step is a single step_interval
    self.max_partial_step = self.max_steps

    # flag to indicate if policy evaluation has converged
    self.convergence = False

    # helper function to display grid information
    def get_info_string( iteration, values, directions=None ):
      info = {'text': values, 'precision': self.precision}

      if directions is not None:
        info['directions'] = {'arrows':directions}

      if iteration is not None:
        info_string = [((10,10),f"Iteration: {iteration}")]
        if self.convergence:
          info_string.append(((10,30),"Convergence"))
        info['side_info'] = info_string
      return info

    def on_update(*args):

      if self.partial_step < self.max_partial_step:

        # do policy updates for the specified number of iterations before a save
        for n in range(self.save_interval):
          self.policy_evaluation.do_iteration()
          self.partial_step += 1

        # calculate the largest difference in the state values from the start to end of the iteration
        delta = np.max(np.abs(self.policy_evaluation.end_values - self.policy_evaluation.start_values))

        # test if the difference is less than the defined convergence threshold
        if delta < self.policy_evaluation.threshold:
          self.convergence = True

        # get and display the state values and directions
        directions = self.policy_evaluation.policy.calculate_greedy_directions(self.policy_evaluation.end_values) if self.show_directions else None
        self.env.show_info(get_info_string(self.partial_step, self.policy_evaluation.end_values, directions))

        # add an extra delay between each step if images are being generated
        if self.create_images: sleep(2)
        self.env.render()

        if ((self.partial_step == self.max_partial_step) or self.convergence):
          self.done = True

          # reduce the max steps in case convergence has finished early
          self.max_partial_step = self.partial_step

    # add the initial state values to the grid at partial_step = 0
    directions = self.policy_evaluation.policy.calculate_greedy_directions(self.policy_evaluation.end_values) if self.show_directions else None
    self.env.show_info(get_info_string(self.partial_step, self.policy_evaluation.end_values, directions))
    self.env.render()

    return self.show(on_update,**kwargs)


import numpy as np

from .baby_robot_env_v5 import BabyRobotEnv_v5
from .lib.direction import Direction
from .lib.actions import Actions


class BabyRobotEnv_v6( BabyRobotEnv_v5 ):
  
  def __init__(self, **kwargs):
      super().__init__(**kwargs) 

  
  def take_action(self, action):
      ''' apply the supplied action 

          returns:
          - the reward obtained for taking the action
          - a flag to indicate if the target state was reached - if it wasn't this indicates
            that a slip has occurred
      '''      

      # convert the action into a direction bitfield
      direction = Direction.from_action(action) 
        
      # calculate the postion of the next state and the reward for moving there
      next_pos,reward,target_reached = self.level.get_next_state( self.x, self.y, direction )  

      # store the new position
      self.x = next_pos[0]
      self.y = next_pos[1]
  
      # update the available actions for the new position        
      self.set_available_actions()      

      return reward, target_reached  


  def step(self, action): 

      # take the action and update the position
      reward, target_reached = self.take_action(action)      
      obs = np.array([self.x,self.y])      
           
      # set the 'terminated' flag if we've reached the exit
      terminated = (self.x == self.end[0]) and (self.y == self.end[1])
      truncated = False
        
      info = {'target_reached':target_reached}
      return obs, reward, terminated, truncated, info            


import numpy as np
import gymnasium
from gymnasium.spaces import Discrete, MultiDiscrete


class BabyRobotEnv_v1(gymnasium.Env):

    def __init__(self, **kwargs):
        super().__init__()

        # dimensions of the grid
        self.width = kwargs.get('width',3)
        self.height = kwargs.get('height',3)

        # define the maximum x and y values
        self.max_x = self.width - 1
        self.max_y = self.height - 1

        # there are 5 possible actions: move N,E,S,W or stay in same state
        self.action_space = Discrete(5)

        # the observation will be the coordinates of Baby Robot
        self.observation_space = MultiDiscrete([self.width, self.height])

        # Baby Robot's position in the grid
        self.x = 0
        self.y = 0

    def step(self, action):
        obs = np.array([self.x,self.y])
        reward = -1
        terminated = True
        truncated = False
        info = {}
        return obs, reward, terminated, truncated, info      

    def reset(self, seed=None, return_info=False, options=None):
        super().reset(seed=seed)
        # reset Baby Robot's position in the grid
        self.x = 0
        self.y = 0
        info = {}
        return np.array([self.x,self.y]),info

    def render(self, info=None ):
        pass


# # Copyright (c) Steve Roberts
# # Distributed under the terms of the Modified BSD License.

# import numpy as np
# from tqdm import tqdm
# from enum import IntEnum
# import babyrobot
# from ..envs import BabyRobotInterface
# from ..envs.lib import Actions
# from ..lib import Policy


# class DeltaType(IntEnum):
#     Max, Mean = range(2)



# class MonteCarlo():

#   last_start_pos = None         # the episode start position
#   rewards: np.array             # rewards array to be initialised by child class
#   values: np.array              # values array to be initialised by child class

#   def __init__(self, policy: Policy, **setup):

#     # process and store any supplied setup parameters
#     self.set_parameters(**setup)

#     # store the policy used to select actions in the environment
#     self.policy = policy

#     # create an evaluation version of the environment
#     self.env = babyrobot.make("BabyRobot-v0", render_mode=None, **setup)


#   def set_parameters(self, **setup):
#     '''
#         set all the parameters using any supplied values,
#         otherwise use the defaults
#     '''
#     # store the setup used to define the environment
#     self.setup = setup
#     self.exploring_starts = setup.get('exploring_starts',False)


#   #
#   # Graphical Helper Functions
#   #

#   def get_graphical_environment(self, text_data):
#     ''' create a graphical version of the environment - virtual base method '''
#     raise NotImplementedError()

#   def show_state_values(self):
#     ''' show the calculated state values '''
#     env = self.get_graphical_environment(self.returns)
#     return env

#   def show_visits(self):
#     ''' show the visits to each state '''
#     env = self.get_graphical_environment(self.visits)
#     return env


#   #
#   # Monte Carlo Method Functions
#   #

#   def get_next_start_state(self):
#     ''' get the environment starting position for the next episode '''

#     # by default begin at the defined start position
#     x = self.env.start[0]
#     y = self.env.start[1]

#     if self.exploring_starts:

#       if self.last_start_pos is not None:
#         # move horizontally to the next state
#         x = self.last_start_pos[0] + 1
#         y = self.last_start_pos[1]

#       while self.env.level.is_grid_state(x,y) == False:
#         # move horizontally to the next state
#         x += 1

#         if x >= self.env.width:
#           # move to the start of the next row
#           x = 0
#           y += 1

#         # test if the last row is complete
#         if y >= self.env.height:
#           # go back to the start
#           x = self.env.start[0]
#           y = self.env.start[1]

#     self.last_start_pos = [x,y]
#     return self.last_start_pos


#   # def rewards_to_returns(self, state_rewards):
#   #   ''' work backwards to convert the rewards into returns '''
#   #   γ = 1.0
#   #   G = 0
#   #   state_returns = []
#   #   for state, reward in reversed(state_rewards):
#   #     G = reward + γ*G
#   #     state_returns.append((state,G))

#   #   # put back into the order of states visited
#   #   state_returns.reverse()
#   #   return state_returns


#   def single_episode(self):
#     '''
#       run a single episode to collect the reward for each action of the trajectory
#       - virtual base method
#     '''
#     raise NotImplementedError()
#     # state,info = self.env.reset()
#     # state_rewards = []
#     # total_reward = 0
#     # terminated = False
#     # while not terminated:
#     #   # get the policy's action in the current state
#     #   action = self.policy.get_action(self.env.x,self.env.y)

#     #   new_state, reward, terminated, truncated, info = self.env.step(action)
#     #   total_reward += reward
#     #   state_rewards.append((state,reward))
#     #   state = new_state

#     # return state_rewards


#   def get_returns(self):
#     raise NotImplementedError()


#   def run(self, max_episodes = 1, delta_interval = 10, delta_type = DeltaType.Max):

#     deltas = []
#     for episode in (pbar :=tqdm(range(max_episodes))):

#       initial_values = self.returns.copy()

#       # set the next start position
#       self.env.set_initial_pos(self.get_next_start_state())

#       rewards = self.single_episode()
#       returns = self.rewards_to_returns(rewards)
#       self.get_returns(returns)

#       # save the change in the state values at regular intervals
#       if episode%delta_interval == 0:
#         # calculate the difference in the state values from the start to end of the iteration
#         if delta_type == DeltaType.Max:
#           delta = np.max(np.abs(self.returns - initial_values))  # get the largest difference
#         else:
#           delta = np.mean(np.abs(self.returns - initial_values)) # get the average difference
#         deltas.append(delta)
#         pbar.set_description(f"Delta {delta:0.5f}")

#     return np.round(self.returns), self.visits, deltas



# class MonteCarloStateValues( MonteCarlo ):
#   ''' base class for Monte Carlo methods calculating state values '''

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)

#     # keep a count of the visits to each state
#     self.visits = np.zeros((self.env.height,self.env.width))

#     # the average returns for each state
#     self.returns = np.zeros((self.env.height,self.env.width))


#   def get_graphical_environment(self, text_data):
#     ''' create a graphical version of the environment '''
#     env = babyrobot.make("BabyRobot-v0", **self.setup)
#     info = {'text': text_data, 'precision': 0}
#     env.show_info(info)
#     return env


#   def rewards_to_returns(self, state_rewards):
#     ''' work backwards to convert the rewards into returns '''
#     γ = 1.0
#     G = 0
#     state_returns = []
#     for state, reward in reversed(state_rewards):
#       G = reward + γ*G
#       state_returns.append((state,G))

#     # put back into the order of states visited
#     state_returns.reverse()
#     return state_returns


#   def single_episode(self):
#     '''
#       run a single episode to collect the reward for each action of the trajectory
#     '''
#     state,info = self.env.reset()
#     state_rewards = []
#     total_reward = 0
#     terminated = False
#     while not terminated:
#       # get the policy's action in the current state
#       action = self.policy.get_action(self.env.x,self.env.y)

#       new_state, reward, terminated, truncated, info = self.env.step(action)
#       total_reward += reward
#       state_rewards.append((state,reward))
#       state = new_state

#     return state_rewards



# class MonteCarloActionValues( MonteCarlo ):
#   ''' base class for Monte Carlo methods calculating action values '''

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)

#     # keep a count of the visits to each action
#     self.visits = np.zeros((self.env.height,self.env.width, len(Actions)))

#     # the average returns for each action
#     self.returns = np.zeros((self.env.height, self.env.width, len(Actions)))


#   def get_graphical_environment(self, text_data):
#     ''' create a graphical version of the environment '''
#     env = babyrobot.make("BabyRobot-v0", **self.setup)
#     info = {'values': text_data, 'precision': 0}
#     env.show_info(info)
#     return env


#   def rewards_to_returns(self, rewards):
#     ''' work backwards to convert the rewards into returns '''
#     γ = 1.0
#     G = 0
#     returns = []
#     for state, action, reward in reversed(rewards):
#       G = γ*G + reward
#       returns.append((state,action,G))

#     # put back into the order of actions visited
#     returns.reverse()
#     return returns


#   def single_episode(self):
#     '''
#       run a single episode to collect the reward for each action of the trajectory
#     '''
#     state,info = self.env.reset()
#     action_rewards = []
#     total_reward = 0
#     terminated = False
#     while not terminated:
#       # get the policy's action in the current state
#       action = self.policy.get_action(self.env.x,self.env.y)

#       new_state, reward, terminated, truncated, info = self.env.step(action)
#       total_reward += reward
#       action_rewards.append((state,action,reward))
#       state = new_state

#     return action_rewards



# '''
#     First Visit Monte Carlo
# '''


# class MonteCarlo_FirstVisit( MonteCarloStateValues ):

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)


#   def get_returns(self, state_returns):
#     ''' find the value of the first visit to a state '''
#     episode_visits = np.zeros((self.env.height,self.env.width))

#     # the first time a state is seen store the return value
#     for s, G in state_returns:
#       # only process the state if it hasn't already been visited
#       if episode_visits[s[1],s[0]] == 0:
#         episode_visits[s[1],s[0]] += 1

#         # increment the total count of first visits to this state
#         self.visits[s[1],s[0]] += 1

#         # get (1/visits) setting the result of any divide by zero to zero
#         inv_visits = 1 / self.visits[s[1],s[0]]

#         # only update the average returns of the states visited in this episode
#         self.returns[s[1],s[0]] = ((1 - inv_visits)*self.returns[s[1],s[0]]) + (inv_visits*G)


#   # def run(self, max_episodes = 1, delta_interval = 10, delta_type = DeltaType.Max):

#   #   deltas = []
#   #   for episode in (pbar :=tqdm(range(max_episodes))):

#   #     initial_values = self.returns.copy()

#   #     # set the next start position
#   #     self.env.set_initial_pos(self.get_next_start_state())

#   #     state_rewards = self.single_episode()
#   #     state_returns = self.rewards_to_returns(state_rewards)
#   #     self.get_first_visit_return(state_returns)

#   #     # save the change in the state values at regular intervals
#   #     if episode%delta_interval == 0:
#   #       # calculate the difference in the state values from the start to end of the iteration
#   #       if delta_type == DeltaType.Max:
#   #         delta = np.max(np.abs(self.returns - initial_values))  # get the largest difference
#   #       else:
#   #         delta = np.mean(np.abs(self.returns - initial_values)) # get the average difference
#   #       deltas.append(delta)
#   #       pbar.set_description(f"Delta {delta:0.5f}")

#   #   return np.round(self.returns), self.visits, deltas



# '''
#     Every Visit Monte Carlo
# '''


# class MonteCarlo_EveryVisit( MonteCarloStateValues ):

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)


#   def get_returns(self, state_returns):
#     ''' use the return from every state visit to calculate the average return '''

#     # the first time a state is seen store the return value
#     for s, G in state_returns:

#       # increment the count of visits to this state
#       self.visits[s[1],s[0]] += 1

#       # get the inverse of visits to this state
#       inv_visits = 1 / self.visits[s[1],s[0]]

#       # only update the average return of the current state
#       self.returns[s[1],s[0]] = ((1 - inv_visits)*self.returns[s[1],s[0]]) + (inv_visits*G)


#   # def run(self, max_episodes = 1, delta_interval = 10, delta_type = DeltaType.Max):

#   #   deltas = []
#   #   for episode in (pbar :=tqdm(range(max_episodes))):

#   #     initial_values = self.returns.copy()

#   #     # set the next start position
#   #     self.env.set_initial_pos(self.get_next_start_state())

#   #     state_rewards = self.single_episode()
#   #     state_returns = self.rewards_to_returns(state_rewards)
#   #     self.get_every_visit_return(state_returns)

#   #     # save the change in the state values at regular intervals
#   #     if episode%delta_interval == 0:
#   #       # calculate the difference in the state values from the start to end of the iteration
#   #       if delta_type == DeltaType.Max:
#   #         delta = np.max(np.abs(self.returns - initial_values))  # get the largest difference
#   #       else:
#   #         delta = np.mean(np.abs(self.returns - initial_values)) # get the average difference
#   #       deltas.append(delta)
#   #       pbar.set_description(f"Delta {delta:0.5f}")

#   #   return np.round(self.returns), self.visits, deltas




# '''
#     First-Visit MC Action Values
# '''

# class MonteCarlo_FirstVisit_ActionValues( MonteCarloActionValues ):

#   def __init__(self, policy: Policy, **setup):
#     super().__init__(policy, **setup)


#   # def single_episode(self):
#   #   '''
#   #     run a single episode to collect the reward for each action of the trajectory
#   #   '''
#   #   state,info = self.env.reset()
#   #   action_rewards = []
#   #   total_reward = 0
#   #   terminated = False
#   #   while not terminated:
#   #     # get the policy's action in the current state
#   #     action = self.policy.get_action(self.env.x,self.env.y)

#   #     new_state, reward, terminated, truncated, info = self.env.step(action)
#   #     total_reward += reward
#   #     action_rewards.append((state,action,reward))
#   #     state = new_state

#   #   return action_rewards


#   # def rewards_to_returns(self, rewards):
#   #   # work backwards to convert the rewards into returns
#   #   γ = 1.0
#   #   G = 0
#   #   returns = []
#   #   for state, action, reward in reversed(rewards):
#   #     G = γ*G + reward
#   #     returns.append((state,action,G))

#   #   # put back into the order of actions visited
#   #   returns.reverse()
#   #   return returns


#   def get_returns(self, returns):
#     ''' find the value of the first visit to a state '''

#     episode_visits = np.zeros((self.env.height, self.env.width, len(Actions)))

#     # the first time a state is seen store the return value
#     for s, a, G in returns:
#       # only process the state action if it hasn't already been visited
#       if episode_visits[s[1],s[0],a] == 0:
#         episode_visits[s[1],s[0],a] += 1

#         # increment the total count of first visits to this state action
#         self.visits[s[1],s[0],a] += 1

#         # get (1/visits) setting the result of any divide by zero to zero
#         inv_visits = 1 / self.visits[s[1],s[0],a]

#         # only update the average returns of the state actions visited in this episode
#         self.returns[s[1],s[0],a] = ((1 - inv_visits)*self.returns[s[1],s[0],a]) + (inv_visits*G)


#   # def run(self, max_episodes = 1, delta_interval = 10, delta_type = DeltaType.Max):

#   #   deltas = []
#   #   for episode in (pbar :=tqdm(range(max_episodes))):

#   #     initial_values = self.returns.copy()

#   #     # set the next start position
#   #     self.env.set_initial_pos(self.get_next_start_state())

#   #     rewards = self.single_episode()
#   #     returns = self.rewards_to_returns(rewards)
#   #     self.get_first_visit_return(returns)

#   #     # save the change in the state values at regular intervals
#   #     if episode%delta_interval == 0:
#   #       # calculate the difference in the state values from the start to end of the iteration
#   #       if delta_type == DeltaType.Max:
#   #         delta = np.max(np.abs(self.returns - initial_values))  # get the largest difference
#   #       else:
#   #         delta = np.mean(np.abs(self.returns - initial_values)) # get the average difference
#   #       deltas.append(delta)
#   #       pbar.set_description(f"Delta {delta:0.5f}")

#   #   return np.round(self.returns), self.visits, deltas


from .direction import Direction


import logging
logging.basicConfig(level=logging.DEBUG, format='(%(threadName)-9s) %(message)s',)


''' control robot positioning
    - non-graphical (should be base class for RobotPosition)
'''
class Robot():

    move_count = 0
    maze = None

    def __init__( self, level, **kwargs ):

      self.level = level.grid_base

      if hasattr(level, 'maze'):
        self.maze = level.maze

      # the position in grid cells
      self.x_cell = 0
      self.y_cell = 0

      # baby robot's initial position
      self.initial_position = kwargs.get('initial_pos',self.level.start)
      self.set_cell_position(self.initial_position)


    def get_cell_position(self):
      ''' get the current position in grid coords '''
      return self.x_cell, self.y_cell


    def set_cell_position(self, *args):
      ''' set the robot position in grid coords '''
      # convert from grid position to pixels
      if len(args) == 1:
        self.x_cell = args[0][0]
        self.y_cell = args[0][1]
      elif len(args) == 2:
        self.x_cell = args[0]
        self.y_cell = args[1]


    def move(self,new_x,new_y):
        ' move from the current position to the specified position '

        while (self.x_cell < (self.level.width-1)) and (self.x_cell < new_x):
          self.x_cell += 1

        while (self.x_cell > 0) and (self.x_cell > new_x):
          self.x_cell -= 1

        while (self.y_cell < (self.level.height-1)) and (self.y_cell < new_y):
          self.y_cell += 1

        while (self.y_cell > 0) and (self.y_cell > new_y):
          self.y_cell -= 1


    def test_for_valid_move( self, direction ):
        ''' test if a move can actually be made in the specified direction '''

        if direction == Direction.Stay:
          return False

        cell = None
        if self.maze is not None:
            x, y = self.get_cell_position()
            cell = self.maze.cell_at( x, y )
            if direction == Direction.North and cell.walls['N']: return False
            if direction == Direction.South and cell.walls['S']: return False
            if direction == Direction.East and cell.walls['E']: return False
            if direction == Direction.West and cell.walls['W']: return False

        return True


# Copyright (c) Steve Roberts
# Distributed under the terms of the Modified BSD License.

import gymnasium
import numpy as np
import random

from .lib.grid_level import GridLevel
from .lib.graphical_grid_level import GraphicalGridLevel
from .lib.robot import Robot
from .lib.robot_draw import RobotDraw
from .lib.dynamic_space import Dynamic
from .lib.direction import Direction
from .lib.actions import Actions


class BabyRobotInterface(gymnasium.Env):
    ''' Baby Robot Gym Environment Base Class '''

    def __init__(self, **kwargs):
        super().__init__()

        # initialize the random seed to allow exact recreation of levels
        self.set_seed(kwargs.get('seed',None))

        # get the rendering mode
        self.render_mode = kwargs.get('render_mode',None)

        # initially no actions are available      
        self.dynamic_action_space = Dynamic()          

        # dimensions of the grid
        self.width = kwargs.get('width',3)
        self.height = kwargs.get('height',3)      
      
        # define the maximum x and y values
        self.max_x = self.width - 1
        self.max_y = self.height - 1

        # the start and end positions in the grid
        # - by default these are the top-left and bottom-right respectively
        self.start = kwargs.get('start',[0,0])       
        self.end = kwargs.get('end',[self.max_x,self.max_y])        
        
        # Baby Robot's initial position and current position in the grid
        # - by default this is the grid start         
        self.set_initial_pos( kwargs.get('initial_pos',self.start) )     
        
        # creation of the level
        if self.render_mode is None:        
          self.level = GridLevel( **kwargs )           
          self.robot = Robot(self.level,**kwargs)   
        else:
          # graphical creation of the level
          self.level = GraphicalGridLevel( **kwargs )           
          self.robot = RobotDraw(self.level,**kwargs)   
          self.robot.draw()                   

        # set the initial position and available actions
        self.reset()


    #
    # Helper Methods
    #   
     
    def set_seed(self, seed = None):
      ''' initialize the random seed to allow exact recreation of levels '''
      if seed is not None:        
        random.seed(seed)
        np.random.seed(seed=seed)

    def take_action(self, action):
        ''' apply the supplied action 

            returns:
            - the reward obtained for taking the action
            - a flag to indicate if the target state was reached - if it wasn't this indicates
              that a slip has occurred
        '''         

        # convert the action into a direction bitfield 
        direction = Direction.from_action(action)  
          
        # calculate the postion of the next state and the reward for moving there
        next_pos,reward,target_reached = self.level.get_next_state( self.x, self.y, direction )  

        # store the new position
        self.x = next_pos[0]
        self.y = next_pos[1]
    
        # update the available actions for the new position        
        self.set_available_actions()      
        return reward, target_reached  


    def get_available_actions( self, x = None, y = None ):
        ''' test which actions are allowed at the specified grid state '''

        # if no coordinate supplied use the current position
        if x is None: x = self.x
        if y is None: y = self.y

        # get the available actions from the grid level
        direction_value = self.level.get_directions(x,y) 

        # convert the grid directions into environment actions
        return Direction.get_action_list(direction_value)  

              
    def set_available_actions( self ):
        ' set the list of available actions into the action space '
        action_list = self.get_available_actions()   
        self.dynamic_action_space.set_actions( action_list )      


    def show_available_actions( self ):
        ''' return a string of avaiable actions for current state '''
        available_actions = str(self.dynamic_action_space.get_available_actions()).replace("'","")
        return f"({self.x},{self.y}) {available_actions:36}"


    def get_transition_probability( self, x = None, y = None, direction: Direction = None ):
        ''' get the probability of moving to the intended target when in the specified cell '''
        # if no coordinate supplied use the current position
        if not x: x = self.x
        if not y: y = self.y      
        probability, barrier = self.level.grid_base.get_transition_probability( x, y, direction )
        return probability, barrier
    
    def get_action_probabilities( self, x, y, action: Actions ):
       ''' for an action in a state, return the list of possible next states and rewards 
           and the accompanying probability for each = p(s',r|s,a)
       '''
       return self.level.get_action_probabilities( x, y, action )        


    def get_reward( self, x, y, direction = None ):
        ''' get the reward for moving to cell (x,y) or, if a direction is specified, 
            the reward for moving from (x,y) to the cell in the specified direction
        '''
        return self.level.get_reward(x,y,direction)

    def set_initial_pos( self, pos ):
        ''' set the initial position of Baby Robot ''' 
        self.initial_pos = pos 
        
        # Baby Robot's current position in the grid                
        self.x = self.initial_pos[0]
        self.y = self.initial_pos[1]                

    #
    # Information Methods
    #    

    def show_info(self,info):
        ''' display the supplied information on the grid level '''
        if self.render_mode is not None:
          self.level.show_info( info )

    def clear_info(self,all_info=False):
        ''' clear any current information of the grid level '''
        if self.render_mode is not None:
          self.level.clear(all_info)
        
    def save(self, filename):
        ''' save the level as an image to the specified file '''
        if self.render_mode is not None:
          self.level.save(filename)                           
